/* Copyright (C) 2018 Chang Wei Tan, Francois Petitjean, Geoff Webb
 This file is part of FastEE.
 FastEE is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, version 3 of the License.
 LbEnhanced is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with LbEnhanced.  If not, see <http://www.gnu.org/licenses/>. */
package timeseriesweka.elasticDistances;

import weka.core.Instance;
import weka.core.Instances;

/**
 * Code for the paper "FastEE: Fast Ensembles of Elastic Distances for Time Series Classification"
 *
 * @author Chang Wei Tan, Francois Petitjean, Geoff Webb
 *
 * Original code from http://www.timeseriesclassification.com/code.php
 * https://bitbucket.org/TonyBagnall/time-series-classification/src/f47c300b937af1d06aac86bdbe228cb9a9d5e00d?at=default
 *
 * MSM distance
 */
public class MSM extends ElasticDistances {
    private final static double[][] matrixD = new double[MAX_SEQ_LENGTH][MAX_SEQ_LENGTH];
    public static double[] msmParams = {
            // <editor-fold defaultstate="collapsed" desc="hidden for space">
            0.01,
            0.01375,
            0.0175,
            0.02125,
            0.025,
            0.02875,
            0.0325,
            0.03625,
            0.04,
            0.04375,
            0.0475,
            0.05125,
            0.055,
            0.05875,
            0.0625,
            0.06625,
            0.07,
            0.07375,
            0.0775,
            0.08125,
            0.085,
            0.08875,
            0.0925,
            0.09625,
            0.1,
            0.136,
            0.172,
            0.208,
            0.244,
            0.28,
            0.316,
            0.352,
            0.388,
            0.424,
            0.46,
            0.496,
            0.532,
            0.568,
            0.604,
            0.64,
            0.676,
            0.712,
            0.748,
            0.784,
            0.82,
            0.856,
            0.892,
            0.928,
            0.964,
            1,
            1.36,
            1.72,
            2.08,
            2.44,
            2.8,
            3.16,
            3.52,
            3.88,
            4.24,
            4.6,
            4.96,
            5.32,
            5.68,
            6.04,
            6.4,
            6.76,
            7.12,
            7.48,
            7.84,
            8.2,
            8.56,
            8.92,
            9.28,
            9.64,
            10,
            13.6,
            17.2,
            20.8,
            24.4,
            28,
            31.6,
            35.2,
            38.8,
            42.4,
            46,
            49.6,
            53.2,
            56.8,
            60.4,
            64,
            67.6,
            71.2,
            74.8,
            78.4,
            82,
            85.6,
            89.2,
            92.8,
            96.4,
            100// </editor-fold>
    };
    private static double c = 0;

    public MSM() {
    }

    public MSM(final double cc) {
        c = cc;
    }

    @Override
    public final double distance(final Instance first, final Instance second) {
        return distance1(first, second, c);
    }

    @Override
    public final double distance(final Instance first, final Instance second, final double cutOffValue) {
        return distance1(first, second, c, cutOffValue);
    }

    public static double distance1(final Instance first, final Instance second, final double c) {
        final int m = first.numAttributes() - 1;
        final int n = second.numAttributes() - 1;
        int i, j;
        double d1, d2, d3;

        // Initialization
        matrixD[0][0] = Math.abs(first.value(0) - second.value(0));
        for (i = 1; i < m; i++) {
            matrixD[i][0] = matrixD[i - 1][0] + editCost(first.value(i), first.value(i - 1), second.value(0), c);
        }
        for (i = 1; i < n; i++) {
            matrixD[0][i] = matrixD[0][i - 1] + editCost(second.value(i), first.value(0), second.value(i - 1), c);
        }

        // Main Loop
        for (i = 1; i < m; i++) {
            for (j = 1; j < n; j++) {
                d1 = matrixD[i - 1][j - 1] + Math.abs(first.value(i) - second.value(j));
                d2 = matrixD[i - 1][j] + editCost(first.value(i), first.value(i - 1), second.value(j), c);
                d3 = matrixD[i][j - 1] + editCost(second.value(j), first.value(i), second.value(j - 1), c);
                matrixD[i][j] = Math.min(d1, Math.min(d2, d3));
            }
        }
        // Output
        return matrixD[m - 1][n - 1];
    }

    public static double distance1(final Instance first, final Instance second, final double c, final double cutOffValue) {
        final int m = first.numAttributes() - 1;
        final int n = second.numAttributes() - 1;
        int i, j;
        double d1, d2, d3;
        double min;

        // Initialization
        matrixD[0][0] = Math.abs(first.value(0) - second.value(0));
        for (i = 1; i < m; i++) {
            matrixD[i][0] = matrixD[i - 1][0] + editCost(first.value(i), first.value(i - 1), second.value(0), c);
        }
        for (i = 1; i < n; i++) {
            matrixD[0][i] = matrixD[0][i - 1] + editCost(second.value(i), first.value(0), second.value(i - 1), c);
        }

        // Main Loop
        for (i = 1; i < m; i++) {
            min = cutOffValue;
            for (j = 1; j < n; j++) {
                d1 = matrixD[i - 1][j - 1] + Math.abs(first.value(i) - second.value(j));
                d2 = matrixD[i - 1][j] + editCost(first.value(i), first.value(i - 1), second.value(j), c);
                d3 = matrixD[i][j - 1] + editCost(second.value(j), first.value(i), second.value(j - 1), c);
                matrixD[i][j] = Math.min(d1, Math.min(d2, d3));

                if (matrixD[i][j] >= cutOffValue) {
                    matrixD[i][j] = Double.MAX_VALUE;
                }

                if (matrixD[i][j] < min) {
                    min = matrixD[i][j];
                }
            }
            if (min >= cutOffValue) {
                return Double.MAX_VALUE;
            }
        }
        // Output
        return matrixD[m - 1][n - 1];
    }

    private static double editCost(final double new_point, final double x, final double y, final double c) {
        double dist;

        if (((x <= new_point) && (new_point <= y)) || ((y <= new_point) && (new_point <= x))) {
            dist = c;
        } else {
            dist = c + Math.min(Math.abs(new_point - x), Math.abs(new_point - y));
        }

        return dist;
    }

    @Override
    public final String toString() {
        return "MSM";
    }

    @Override
    public final void setParamsFromParamID(final Instances train, final int paramId) {
        c = msmParams[paramId];
    }

    public final String getParams() {
        return "c = " + c;
    }

    public final double getC() {
        return c;
    }
}
